'''Tests for the to_json() method.'''

from sqlalchemy import select
from sqlalchemy.orm import joinedload, load_only
from .models import *
import json

def test_to_json_single(session):

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    expected = '''{"user": {"id": 1, "firstname": "test"}}'''

    test = User.to_json(user, fields=('id', 'firstname'),
            relationships=False)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)


def test_to_json_multiple(session):

    users = session.scalars(
        select(User).where(User.id.in_((1, 2)))
    )

    expected = '''{"users": [{"id": 1, "firstname": "test"},{"id": 2, "firstname": "test"}]}'''

    test = User.to_json(users, fields=('id', 'firstname'),
            relationships=False)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)


def test_to_json_empty(session):

    users = session.scalars(
        select(User).where(User.id.in_((345432, 3412)))
    )
    expected = '''{"users": []}'''

    test = User.to_json(users, fields=('id', 'firstname'),
            relationships=False)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)


def test_to_json_none(session):

    user = session.scalars(
        select(User).where(User.id == 52354)
    ).first()

    expected = '''{"user": null}'''

    test = User.to_json(user, fields=('id', 'firstname'),
            relationships=False)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)

def test_to_json_blob(session):

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    expected = '''{"user": {"id": 1, "blob": "AQIDBA=="}}'''

    test = User.to_json(user, fields=('id', 'blob'),
            relationships=False)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)



'''Tests for the async_factory() method.'''

from sqlalchemy import select
from sqlalchemy.orm import load_only
from .models import *
import json

import pytest

@pytest.mark.asyncio
async def test_async_factory_single(session):

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    expected = b'''{"user": {"id": 1, "firstname": "test"}}'''

    generator = await User.async_factory(user, fields=('id', 'firstname'),
            relationships=False)

    items = []
    async for j in generator():
        items.append(j)

    test = b''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)

@pytest.mark.asyncio
async def test_async_factory_multiple(session):

    # users = User.query.filter(User.id.in_((1, 2)))
    users = session.scalars(
        select(User)
        .options(load_only(User.id, User.firstname))
        .where(User.id.in_((1, 2)))
    )

    expected = b'''{"users": [{"id": 1, "firstname": "test"},{"id": 2, "firstname": "test"}]}'''

    generator = await User.async_factory(users, fields=('id', 'firstname'),
            relationships=False)

    items = []
    async for j in generator():
        items.append(j)

    test = b''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)

@pytest.mark.asyncio
async def test_async_factory_empty(session):

    # users = User.query.filter(User.id.in_((345432, 3412)))
    users = session.scalars(
        select(User).where(User.id.in_((345432, 3412)))
    )

    expected = b'''{"users": []}'''

    generator = await User.async_factory(users, fields=('id', 'firstname'),
            relationships=False)

    items = []
    async for j in generator():
        items.append(j)

    test = b''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)

@pytest.mark.asyncio
async def test_async_factory_none(session):

    # user = User.query.\
    #     filter(User.id == 52354).\
    #     first()
    user = session.scalars(
        select(User).where(User.id == 52354)
    ).first()

    expected = b'''{"user": null}'''

    generator = await User.async_factory(user, fields=('id', 'firstname'),
            relationships=False)

    items = []
    async for j in generator():
        items.append(j)

    test = b''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)


@pytest.mark.asyncio
async def test_async_factory_single_str(session):

    # user = User.query.\
    #     filter(User.id == 1).\
    #     one()
    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    expected = '''{"user": {"id": 1, "firstname": "test"}}'''

    generator = await User.async_factory(user, fields=('id', 'firstname'),
            relationships=False, encoding=False)

    items = []
    async for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)

@pytest.mark.asyncio
async def test_async_factory_multiple_str(session):

    # users = User.query.filter(User.id.in_((1, 2)))
    users = session.scalars(
        select(User).where(User.id.in_((1, 2)))
    )

    expected = '''{"users": [{"id": 1, "firstname": "test"},{"id": 2, "firstname": "test"}]}'''

    generator = await User.async_factory(users, fields=('id', 'firstname'),
            relationships=False, encoding=False)

    items = []
    async for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)

@pytest.mark.asyncio
async def test_async_factory_empty_str(session):

    # users = User.query.filter(User.id.in_((345432, 3412)))
    users = session.scalars(
        select(User).where(User.id.in_((345432, 3412)))
    )

    expected = '''{"users": []}'''

    generator = await User.async_factory(users, fields=('id', 'firstname'),
            relationships=False, encoding=False)

    items = []
    async for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)

@pytest.mark.asyncio
async def test_async_factory_none_str(session):

    # user = User.query.\
    #     filter(User.id == 52354).\
    #     first()
    user = session.scalars(
        select(User).where(User.id == 52354)
    ).first()

    expected = '''{"user": null}'''

    generator = await User.async_factory(user, fields=('id', 'firstname'),
            relationships=False, encoding=False)

    items = []
    async for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)




import sqlalchemy as sa
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declarative_base
from datetime import datetime, date, time, timedelta
import json

from serialalchemy import Serializable

from models import *


def populate_test_data(session):
    for i in range(5, 45):
        u = User()

        u.firstname = "test"
        u.lastname = "test"

        u.number = 21.3534
        u.flt = 1.3432
        u.created = datetime(2015, 8, 31, 15, 28, 33)

        u.pickled = ['one', 'two']
        u.selections = 'one'
        u.something = False

        u.birthdate = date(2015, 9, 5)
        u.alarm = time(15, 28, 33)
        u.interval = timedelta(days=-10)

        u.username = 'uname'

        u.data = {
                'id': 'fdas',
                'value': 'Lorem ipsum',
                children: [
                    {'id': 'fjdiosjo', 'title': 'jiobjvjifog'},
                    {'id': 'fjsdio', 'title': 'fjnan'},
                    {'id': 'ajdoa', 'title': '3u490j3fj'},
                    ]
                }


        p = Profile()
        p.somefield = 'somevalue'
        u.profile = p

        g1 = Group()
        g1.id = i+1
        g1.name = 'group1'

        g2 = Group()
        g2.id = i + 45
        g2.name = 'group2'

        u.groups[:] = [g1, g2]

        session.add(u)
        session.add_all([g1, g2])

    session.commit()

    p = [
         Page(title='Test Page', content='test'),
         Page(title='Test Page', content='test'),
         Page(title='Test Page', content='test'),
         Page(title='Test Page', content='test'),
         ]

    session.add_all(p)
    session.commit()

    p1 = Page.query.get(1)
    p2 = Page.query.get(2)

    p2.parent = p1
    session.commit()



engine = sa.create_engine('sqlite:///:memory:', echo=False)
Session.configure(bind=engine)

session = Session()

Base.metadata.create_all(engine)

populate_test_data(session)



'''Tests for the json_factory() method.'''

from sqlalchemy import select
from .models import *
import json

def test_json_factory_single(session):

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    expected = '''{"user": {"id": 1, "firstname": "test"}}'''

    generator = User.json_factory(user, fields=('id', 'firstname'),
            relationships=False)

    items = []
    for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)


def test_json_factory_multiple(session):

    users = session.scalars(
        select(User).where(User.id.in_((1, 2)))
    )

    expected = '''{"users": [{"id": 1, "firstname": "test"},{"id": 2, "firstname": "test"}]}'''

    generator = User.json_factory(users, fields=('id', 'firstname'),
            relationships=False)

    items = []
    for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)


def test_json_factory_empty(session):

    users = session.scalars(
        select(User).where(User.id.in_((345432, 3412)))
    )

    expected = '''{"users": []}'''

    generator = User.json_factory(users, fields=('id', 'firstname'),
            relationships=False)

    items = []
    for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)


def test_json_factory_none(session):

    user = session.scalars(
        select(User).where(User.id == 23213412)
    ).first()

    expected = '''{"user": null}'''

    generator = User.json_factory(user, fields=('id', 'firstname'),
            relationships=False)

    items = []
    for j in generator():
        items.append(j)

    test = ''.join(items)

    de = json.loads(expected)
    dt = json.loads(test)

    assert json.dumps(de) == json.dumps(dt)




'''Tests for to_dict() method'''

from .models import *
from sqlalchemy.orm import joinedload, load_only
from sqlalchemy import select

def compare_dicts(expected, test):
    '''recursive dict comparison'''

    keys = set(expected.keys()) | set(test.keys())

    for key in keys:
        if isinstance(expected[key], dict) and isinstance(test[key], dict):
            compare_dicts(expected[key], test[key])
        else:
            assert test[key] == expected[key]



def test_plain(session):
    '''Serialize single object, no relationships'''
    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict(relationships=False)

    expected = {
            'id': 1,
            'firstname': 'test',
            'lastname': 'test',
            'number': 21.3534,
            'flt': 1.3432,
            'created': '2015-08-31 15:28:33',
            'pickled': ['one', 'two'],
            'selections': 'one',
            'something': False,
            'birthdate': '2015-09-05',
            'alarm': '15:28:33',
            'interval': -864000.0,
            'username': 'uname',
            'two': 2,
            'password': None,
            'smalltext': None,
            'four': 4,
            'blob': b'\x01\x02\x03\x04'
            }

    compare_dicts(expected, data)

def test_fields(session):
    '''Serialize single object, no relationships, limiting fields'''
    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict(fields=('firstname', 'lastname'), relationships=False)

    expected = {
            'firstname': 'test',
            'lastname': 'test',
            }

    compare_dicts(expected, data)

def test_fields2(session):
    '''Serialize single object, no relationships, limiting fields, this time
    with properties and non-existent fields'''
    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict(fields=('firstname', 'lastname', 'two', 'nonexistent'), relationships=False)

    expected = {
            'firstname': 'test',
            'lastname': 'test',
            'two': 2
            }

    compare_dicts(expected, data)

def test_fields3(session):
    '''Serialize single object, no relationships, limiting fields, this time
    ignoring properties that are not decorated with serializable_property'''

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict(fields=('firstname', 'lastname', 'two', 'three'), relationships=False)

    expected = {
            'firstname': 'test',
            'lastname': 'test',
            'two': 2
            }

    compare_dicts(expected, data)


def test_fields4(session):
    '''Serialize single object, with relationships, limiting fields, and
        related objects have a serializable_property'''

    stmt = (select(Profile)
            .options(joinedload(Profile.user)
                     .load_only(User.id))
            .where(Profile.id == 1))

    profile = session.scalars(
        stmt
    ).one()

    data = profile.to_dict(fields=('id', 'user'))

    expected = {
            'id': 1,
            'user': {
                'id': 1,
                'two': 2,
                'four': 4,
                },
            }

    compare_dicts(expected, data)


def test_exclude_fields(session):
    '''Serialize single object no relationship, excluding fields'''

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict(fields=['~firstname'], relationships=False)

    assert 'firstname' not in data



def test_relationships(session):
    '''Serialize object with related objects. MTM relationships return PK only
    (default setup)'''

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict()

    expected = {
            'id': 1,
            'firstname': 'test',
            'lastname': 'test',
            'number': 21.3534,
            'flt': 1.3432,
            'created': '2015-08-31 15:28:33',
            'pickled': ['one', 'two'],
            'selections': 'one',
            'something': False,
            'birthdate': '2015-09-05',
            'alarm': '15:28:33',
            'interval': -864000.0,
            'username': 'uname',
            'password': None,
            'smalltext': None,
            'two': 2,
            'profile': {
                'id': 1,
                'user_id': 1,
                'somefield': 'somevalue'
                },
            'groups': [6, 50],
            'four': 4,
            'blob': b'\x01\x02\x03\x04',
            }


    compare_dicts(expected, data)


def test_relationships_fullmtm(session):
    '''Serialize object with related objects. Full MTM objects returned.'''

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict(mtm_pkonly=False)

    expected = {
            'id': 1,
            'firstname': 'test',
            'lastname': 'test',
            'number': 21.3534,
            'flt': 1.3432,
            'created': '2015-08-31 15:28:33',
            'pickled': ['one', 'two'],
            'selections': 'one',
            'something': False,
            'birthdate': '2015-09-05',
            'alarm': '15:28:33',
            'interval': -864000.0,
            'username': 'uname',
            'password': None,
            'two': 2,
            'smalltext': None,
            'four': 4,
            'blob': b'\x01\x02\x03\x04',
            'profile': {
                'id': 1,
                'user_id': 1,
                'somefield': 'somevalue'
                },
            'groups': [
                {'id': 6, 'name': 'group1'},
                {'id': 50, 'name': 'group2'},
                ],
            }

    compare_dicts(expected, data)


def test_relationships_loadonly(session):
    '''Serialize object with related objects, limiting field output on parent
        and child.'''


    stmt = (
        select(User)
        .options(joinedload(User.profile)
                 .load_only(Profile.somefield))
        .where(User.id == 3)
    )
    print(stmt)
    user = session.scalars(
        stmt
    ).one()

    data = user.to_dict(['firstname', 'profile'])


    expected = {
            'firstname': 'test',
            'profile': {
                'id': 3,
                'somefield': 'somevalue',
                },
            }

    compare_dicts(expected, data)


def test_self_ref(session):
    '''Serialize object of a model with a self-reference'''

    page = session.scalars(
        select(Page).where(Page.id == 1)
    ).one()

    data = page.to_dict()

    expected = {
            'id': 1,
            'parent_id': None,
            'title': 'Test Page',
            'content': 'test',
            'children': [
                {
                    'title': 'Test Page',
                    'content': 'test',
                    'parent_id': 1,
                    'id': 2,
                },
                ]
            }

    compare_dicts(expected, data)


def test_self_ref_empty(session):
    '''Serialize object of a model with a self-reference'''

    page = session.scalars(
        select(Page).where(Page.id == 3)
    ).one()

    data = page.to_dict()

    expected = {
            'id': 3,
            'parent_id': None,
            'title': 'Test Page',
            'content': 'test',
            }

    compare_dicts(expected, data)

def test_hybrid_property(session):
    '''Test serialization of hybrid attribute'''

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = user.to_dict(fields=('firstname', 'lastname', 'four'), relationships=False)

    expected = {
            'firstname': 'test',
            'lastname': 'test',
            'four': 4,
            }


    compare_dicts(expected, data)




from models import *
from testdata import *

from sqlalchemy.orm import joinedload

import pprint
import json

user = User.query.get(1)
pprint.pprint(user.to_dict())
'''>>>
{'addresses': [],
 'alarm': '15:28:33',
 'birthdate': '2015-09-05',
 'created': '2015-08-31 15:28:33',
 'firstname': 'test',
 'flt': 1.3432,
 'groups': [6, 50],
 'id': 1,
 'interval': -864000.0,
 'lastname': 'test',
 'number': 21.3534,
 'password': None,
 'pickled': ['one', 'two'],
 'profile': {'id': 1, 'somefield': 'somevalue', 'user_id': 1},
 'selections': 'one',
 'something': False,
 'two': 2,
 'username': 'uname'}
'''

pprint.pprint(user.to_dict(fields=['id', 'firstname', 'lastname']))
'''>>>
{'firstname': 'test', 'id': 1, 'lastname': 'test'}
'''

#SQLAlchemy caches results, so get a different user
user = User.query.\
        options(joinedload(User.profile).load_only('somefield')).\
        get(2)
pprint.pprint(user.to_dict(fields=['id', 'profile']))
'''>>>
{'id': 2, 'profile': {'somefield': 'somevalue'}}
'''

#JSON utility
gen = User.json_factory(user, fields=['id', 'profile'])
for item in gen():
    print(item)
'''>>>
{"user": {"id": 2, "profile": {"somefield": "somevalue"}}}
'''


u = User()
errors = u.populate({
    'smalltext': 'jfiodsajofidiojsfojdsa',
    }, swallow_exceptions=True)

print(errors)



'''SerialAlchemy
Dead simple object serialization for SQLAlchemy

'''

from sqlalchemy import inspect, DateTime, Date, Time, event
from decimal import Decimal
from datetime import datetime, date, timedelta, time
from sqlalchemy.ext.hybrid import hybrid_property

# import inspect as inspectPy

import re
from collections import abc
import json
import base64

from dateutil.parser import (
    parse,
    ParserError,
)



class SerialAlchemyError(Exception):
    pass

__all__ = [
        'serializable_property',
        'serializable_hybrid',
        'Serializable',
        ]

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')
def convert_class(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()



def _get_value(obj, field):
    '''Returns a value that can be serialized.

    2023-07-10: Added docstring because I forgot what this did :P
    '''

    primitives = (int, str, bytes, float, dict, list, type(None),)

    value = getattr(obj, field)

    if isinstance(value, Decimal):
        return float(value)
    elif isinstance(value, datetime) or isinstance(value, date) or\
            isinstance(value, time):
        return str(value)
    elif isinstance(value, timedelta):
        return float(value.total_seconds())
    elif isinstance(value, (set, frozenset,)):
        return list(value)
    elif not isinstance(value, primitives):
        # last ditch effort, try converting to string
        return str(value)

    return value


def _get_serializable_properties(obj, attrs, loaded):
    all_members = set([f for f in dir(obj) if not f.startswith('_')])
    properties = all_members - set(attrs) - set(loaded)

    if not hasattr(obj, '__class__'):
        return []

    cls = obj.__class__
    propout = []

    obj_type = type(obj)

    for prop in properties:
        inst = getattr(obj_type, prop)
        if inst and \
                (isinstance(inst, serializable_property) or\
                 isinstance(inst, serializable_hybrid)):

            propout.append(prop)
        else:
            attr = getattr(cls, prop)
            should_add = False

            try:
                should_add = isinstance(attr.descriptor, hybrid_property)
            except AttributeError:
                should_add = False

            if should_add and hasattr(obj, prop):
                propout.append(prop)

    return propout


class ExtJSONEncoder(json.JSONEncoder):
    ''' Add a way to serialize bytes to base64.
    '''
    def default(self, obj):
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode()

        # return json.JSONEncoder.default(self, obj)
        return super().default(obj)


def _str_to_datetime(value):
    try:
        dt = parse(value)
    except Exception as ex:
        raise SerialAlchemyError('Datetime parsing error', ex)
    return dt

def _str_to_date(value):
    if value is None:
        return None
    try:
        dt = parse(value)
    except Exception as ex:
        raise SerialAlchemyError('Date parsing error', ex)
    return dt.date()

def _str_to_time(value):
    try:
        dt = parse(value)
    except Exception as ex:
        raise SerialAlchemyError('Date parsing error', ex)
    return dt.time()


class serializable_property(property):
    '''Simple wrapper for built-in `property` for explicit serialization of
    class/instance properties.

    As of version 0.2.0, SerialAlchemy will not include a property unless
    it is decorated with `serializable_property`.
    '''
    pass

class serializable_hybrid(hybrid_property):
    '''A wrapper for sqlalchemy's `hybrid_property` for explicit serialization of
    a decorated hybrid property.

    As of version 0.5.0, SerialAlchemy will not include a hybrid property unless
    it is decorated with `serializable_hybrid`.
    '''
    pass


class Serializable:
    '''Provides serialization and deserialization for data-mapped classes.

    This is a mixin class that provides serializaiton to python dicts and
    object population (deserialization). A method for creating a streaming
    json generator from a result set is also defined.

    The mixin defines two special attributes during SQLAlchemy's
    `mapper_configured` event which are used for json generation.

    :attribute __single__: The name of a single object. If not specified,
        the mapper class name is used, converting camel-case to
        underscore.

    :attribute __plural__: The name of multiple objects. If not specified,
        the mapper's `__tablename__` is used.

    This is my own personal preference, although I'm told a few other people do
    it this way too::
        class User(Base, Serializable):
            __tablename__ = 'users'

            #rest of class definition

    In this example, __single__ = 'user', __plural__ = 'users'.

    Properties must be serialized with the `@serializable_property`
    or the `@hybrid_property` decorator to be included in the output. Functions
    decorated with Python's `property` built-in are ignored.::
        class User(Base, Serializable):
            __tablename__ = 'users'

            #Columns...

            @property
            def one(self):
                """this will not be in the output"""
                return 1

            @serializable_property
            def two(self):
                """this will be in the output"""
                return 2

            @serializable_hybrid
            def three(self):
                return self.third_field

            @three.expression
            def three(cls):
                return select(cls.third_field).label('three')


    Additionally, specific columns can be excluded using the column's info
    attribute::
        class User(Base, Serializable):
            __tablename__ = 'users'

            username = Column(Unicode(100))
            password = Column(UnicodeText,
                info={'serializable': False})

    In this example, password will never be included in the output of
    `to_dict()`.
    '''

    @classmethod
    def _define_special_fields(cls):
        if not hasattr(cls, '__single__') or cls.__single__ is None:
            cls.__single__ = convert_class(cls.__name__)

        if not hasattr(cls, '__plural__') or cls.__plural__ is None:
            if hasattr(cls, '__tablename__'):
                cls.__plural__ = cls.__tablename__
            else:
                cls.__plural__ = cls.__single__


    def to_dict(self, fields=None, relationships=True, mtm_pkonly=True):
        '''Serialize a SQLAlchemy mapper in to a python dictionary.

        :param fields=None: An iterable of field names (str) or None.
            The dictionary returned will include only the fields specified.
            Alternatively, prefixing a field name with a tilde (~) will
            exclude that field from the returned dict. Mixing included and
            excluded field names will ignore the excluded fields.

            Relationship properties can be in this parameter, although will be
            ignored unless the `relationships` parameter is True. Properties of
            related objects are filtered via SQLAlchemy's `load_only` query
            option.

        :param relationships=True: Include or omit relationship properties.
            Output is based on the relationship definition. It will be either
            a list of dicts, or a dict itself based on the `uselist` parameter
            of the relationship function.

            The `fields` paramter does not affect the related object output.
            To filter related object fields, use SQLAlchemy's `load_only`
            device during query construction.

            The related class must also have the `Serializable` mixin.

            *Note: Relationships of related objects are ignored. I couldn't
            see the benefit vs the complexity it would bring.*

        :param mtm_pkonly=True: If True, return only the primary-key for
            secondary objects of a many-to-many relationship. If False,
            all fields of the secondary object will be included, subject
            to the `load_only` filter.
        '''

        output = {}

        includes = frozenset()
        excludes = frozenset()

        if fields is not None:
            includes = frozenset([f for f in fields if not f.startswith('~')])
            excludes = frozenset([f[1:] for f in fields if f.startswith('~')])

        info = inspect(self)

        for col in info.mapper.columns:
            field = col.name
            if includes and field not in includes:
                continue
            elif excludes and field in excludes:
                continue
            elif hasattr(col, 'info'):
                if 'serializable' in col.info and not col.info['serializable']:
                    continue

            output[field] = _get_value(self, field)


        if relationships:
            relations = info.mapper.relationships.items()
            for field, relation in relations:
                if includes and field not in includes:
                    continue
                elif excludes and field in excludes:
                    continue

                data = getattr(self, field)
                relout = None

                if relation.uselist:
                    relout = []
                    for obj in data:
                        if isinstance(obj, Serializable):
                            if relation.secondary is not None and mtm_pkonly:
                                relinfo = inspect(obj.__class__)
                                pkeys = relinfo.primary_key
                                out = []
                                for pk in pkeys:
                                    val = getattr(obj, pk.name)
                                    out.append(val)
                                relout.extend(out)
                            else:
                                relinfo = inspect(obj)
                                relfields = [k for k in relinfo.dict.keys() if
                                        not k.startswith('_')]

                                props = _get_serializable_properties(obj,
                                        relinfo.attrs.keys(),
                                        relfields)

                                relfields.extend(props)

                                relout.append(obj.to_dict(relfields,
                                    relationships=False,
                                    mtm_pkonly=mtm_pkonly))

                else:
                    if data and isinstance(data, Serializable):
                        relinfo = inspect(data)
                        relfields = [k for k in relinfo.dict.keys() if not
                                k.startswith('_')]

                        props = _get_serializable_properties(data,
                                relinfo.attrs.keys(), relfields)

                        relfields.extend(props)

                        relout = data.to_dict(relfields,
                                relationships=False,
                                mtm_pkonly=mtm_pkonly)

                if relout:
                    output[field] = relout


        properties = _get_serializable_properties(self, info.attrs.keys(), [])

        for prop in properties:
            if includes and prop not in includes:
                continue
            elif excludes and prop in excludes:
                continue

            if hasattr(self, prop):
                output[prop] = _get_value(self, prop)

        return output

    def populate(self, data, skip_fields=None, swallow_exceptions=False,
                 strict_typing=True):
        '''Populate object from a dict of fields.

        *This method does not populate relationships*. It is recommended
        to override the method and handle them manually.

        Values for relationship attributes can be supplied in the `data`
        parameter and will be safely ignored.

        :param data: A dict whose keys match the object attributes.
        :param skip_fields: An iterable of attributes to not set, whether they
            are in the data parameter or not.
        :param swallow_exceptions: Boolean flag to blanket handle all
            exceptions when setting values, and return a list of Error objects.
            Hint: use with SQLAlchemy's `validates` decorator.
            Default: False
        :param strict_typing: Boolean flag to raise a TypeError exception if
            the incoming value is not of the same type.
            Default: True
        '''

        if skip_fields is None:
            skip_fields = []

        info = inspect(self)
        mapper = info.mapper
        errors = []

        for col in mapper.columns:
            if col.name in skip_fields:
                continue
            elif col.name in data:
                if isinstance(col.type, DateTime):
                    value = _str_to_datetime(data[col.name])
                elif isinstance(col.type, Date):
                    value = _str_to_date(data[col.name])
                elif isinstance(col.type, Time):
                    value = _str_to_time(data[col.name])
                else:
                    value = data[col.name]

                col_pytype = col.type.python_type
                none_ok = col.nullable and value is None
                try:
                    if strict_typing and not isinstance(value, col_pytype) and\
                            not none_ok:
                        raise TypeError(col.name,
                                        f'Value {value} {type(value)} does '\
                                        f'not match type {str(col_pytype)}')
                    setattr(self, col.name, value)
                except Exception as ex:
                    if not swallow_exceptions:
                        raise ex
                    errors.append(ex)

        return errors

    @classmethod
    def dict_factory(cls, resultset, fields=None, relationships=True,
            mtm_pkonly=True):
        '''Create a generator to serialize a result or result set to a dict.

        This classmethod creates a generator that yields each row of a
        resultset, or a single result.

        :param resultset: An iterable (e.g. list or SQLAlchemy query), or a
            single mapped instance.

        :param fields=None: Passed directly to `to_dict`
        :param relationships=True: Passed directly to `to_dict`
        :param mtm_pkonly=True: Passed directly to `to_dict`

        '''

        def generator():
            if isinstance(resultset, abc.Iterable):

                iresult = iter(resultset)
                try:
                    prev = next(iresult)

                    for row in iresult:
                        yield prev.to_dict(fields, relationships, mtm_pkonly)
                        prev = row

                    yield prev.to_dict(fields, relationships, mtm_pkonly)
                except StopIteration:
                    pass
                except Exception as ex:
                    raise ex

            elif resultset is not None:
                yield resultset.to_dict(fields, relationships, mtm_pkonly)
            else:
                yield dict()

        return generator

    @classmethod
    def json_factory(cls, resultset, fields=None, relationships=True,
            mtm_pkonly=True):
        '''Create a generator to serialize a result or result set to JSON.

        This classmethod creates a generator that streams JSON data, wrapping
        the set in the pluralized name if the object is iterable (i.e. a list),
        or in the singular name if not.

        :param resultset: An iterable (e.g. list or SQLAlchemy query), or a
            single mapped instance.

        :param fields=None: Passed directly to `to_dict`
        :param relationships=True: Passed directly to `to_dict`
        :param mtm_pkonly=True: Passed directly to `to_dict`

        '''

        def generator():
            if isinstance(resultset, abc.Iterable):
                yield '{"%s": [' % cls.__plural__

                iresult = iter(resultset)
                try:
                    prev = next(iresult)

                    for row in iresult:
                        yield json.dumps(prev.to_dict(fields, relationships,
                            mtm_pkonly), cls=ExtJSONEncoder) + ','
                        prev = row

                    yield json.dumps(prev.to_dict(fields, relationships,
                        mtm_pkonly), cls=ExtJSONEncoder)
                except StopIteration:
                    pass
                except Exception as ex:
                    raise ex
                finally:
                    yield ']}'

            elif resultset is not None:
                yield json.dumps({cls.__single__: resultset.to_dict(fields,
                    relationships, mtm_pkonly)}, cls=ExtJSONEncoder)
            else:
                yield json.dumps({cls.__single__: None})

        return generator

    @classmethod
    def to_json(cls, resultset, fields=None, relationships=True,
            mtm_pkonly=True):
        '''Dump a result or result set to JSON.

        This classmethod is similar to `json_factory` but does not create a
        generator for streaming output. The results are wrapped in the
        pluralized name if the object is iterable (i.e. a list), or in the
        singular name if not.

        :param resultset: An iterable (e.g. list or SQLAlchemy query), or a
            single mapped instance.

        :param fields=None: Passed directly to `to_dict`
        :param relationships=True: Passed directly to `to_dict`
        :param mtm_pkonly=True: Passed directly to `to_dict`

        '''

        if isinstance(resultset, abc.Iterable):
            seritems = []

            for item in resultset:
                seritems.append(item.to_dict(fields, relationships,
                    mtm_pkonly))

            return json.dumps({cls.__plural__: seritems}, cls=ExtJSONEncoder)
        elif resultset is not None:
            return json.dumps({cls.__single__: resultset.to_dict(fields,
                relationships, mtm_pkonly)}, cls=ExtJSONEncoder)
        else:
            return json.dumps({cls.__single__: None})


    @classmethod
    async def async_factory(cls, resultset, fields=None, relationships=True,
            mtm_pkonly=True, encoding='utf-8'):
        '''Experimental: Create an async generator to serialize a result or result set to JSON.

        This classmethod creates a generator that streams JSON data, wrapping
        the set in the pluralized name if the object is iterable (i.e. a list),
        or in the singular name if not.

        :param resultset: An iterable (e.g. list or SQLAlchemy query), or a
            single mapped instance.

        :param fields=None: Passed directly to `to_dict`
        :param relationships=True: Passed directly to `to_dict`
        :param mtm_pkonly=True: Passed directly to `to_dict`
        :param encoding='utf-8': Encoding for the result. Defaults to 'utf-8'.
            Set to `False` to return an unencoded str.
            (This is mostly because Hypercorn doesn't seem to encode a
            response stream the way Werkzeug does).
        '''

        async def generator():
            if isinstance(resultset, abc.Iterable):
                start = '{"%s": [' % cls.__plural__
                if encoding:
                    start = start.encode(encoding)

                yield start

                iresult = iter(resultset)
                try:
                    prev = next(iresult)

                    for row in iresult:
                        r = json.dumps(prev.to_dict(fields, relationships,
                            mtm_pkonly)) + ','

                        if encoding:
                            r = r.encode(encoding)

                        yield r

                        prev = row

                    p = json.dumps(prev.to_dict(fields, relationships, mtm_pkonly))
                    if encoding:
                        p = p.encode(encoding)

                    yield p
                except StopIteration:
                    pass
                except Exception as ex:
                    raise ex
                finally:
                    end = ']}'
                    if encoding:
                        end = end.encode(encoding)

                    yield end

            elif resultset is not None:
                s = json.dumps({cls.__single__: resultset.to_dict(fields,
                    relationships, mtm_pkonly)})

                if encoding:
                    s = s.encode(encoding)

                yield s
            else:
                empty = json.dumps({cls.__single__: None})
                if encoding:
                    empty = empty.encode(encoding)

                yield empty

        return generator


@event.listens_for(Serializable, 'mapper_configured', propagate=True)
def add_special_fields(mapper, cls):
    cls._define_special_fields()




import sqlalchemy as sa
from sqlalchemy.orm import relationship, backref, validates
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.ext.hybrid import hybrid_property

from serialalchemy import *

from custom_types import *

from typing import NewType


class ValidationError(Exception):
    pass


Base = declarative_base()
session_factory = sa.orm.sessionmaker()
Session = sa.orm.scoped_session(session_factory)
Base.query = Session.query_property()


users_groups = sa.Table('users_groups', Base.metadata,
        sa.Column('user_id', sa.Integer, sa.ForeignKey('users.id',
            onupdate='cascade', ondelete='cascade'), primary_key=True),
        sa.Column('group_id', sa.Integer, sa.ForeignKey('groups.id',
            onupdate='cascade', ondelete='cascade'), primary_key=True)
    )

class User(Base, Serializable):
    __tablename__ = 'users'

    id: int = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    username: str = sa.Column(sa.UnicodeText, nullable=False)
    password: str = sa.Column(sa.UnicodeText)

    firstname = sa.Column(sa.Unicode(100))
    lastname = sa.Column(sa.Unicode(100))

    birthdate = sa.Column(sa.Date)
    created = sa.Column(sa.DateTime)
    alarm = sa.Column(sa.Time)

    pickled = sa.Column(sa.PickleType)
    number = sa.Column(sa.Numeric)
    flt = sa.Column(sa.Float)
    something = sa.Column(sa.Boolean)

    always_excluded = sa.Column(sa.UnicodeText,
            info={'serializable': False})

    selections = sa.Column(sa.Enum('one', 'two', 'three'))
    interval = sa.Column(sa.Interval)

    guid = sa.Column(GUID)

    smalltext = sa.Column(sa.Unicode(5))

    data = sa.Column(sa.JSON)

    @serializable_property
    def two(self):
        print('two executed')
        return 1 + 1

    @hybrid_property
    def test(self):
        return self.number

    @validates('smalltext')
    def valid_name(self, key, value):
        if len(value) > 5:
            raise ValidationError(key, 'Invalid value size')
        return value


class Profile(Base, Serializable):
    __tablename__ = 'profiles'

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    user_id = sa.Column(sa.Integer, sa.ForeignKey('users.id',
        onupdate='cascade', ondelete='cascade'))

    somefield = sa.Column(sa.UnicodeText)

    user = relationship('User', backref=backref('profile',
        uselist=False))


class Address(Base, Serializable):
    __tablename__ = 'addresses'

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    user_id = sa.Column(sa.Integer, sa.ForeignKey('users.id',
        onupdate='cascade', ondelete='cascade'))

    address = sa.Column(sa.UnicodeText)

    user = relationship('User', backref='addresses')


class Group(Base, Serializable):
    __tablename__ = 'groups'

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)

    name = sa.Column(sa.UnicodeText)

    users = relationship('User', secondary=users_groups, backref='groups')


class Compkey(Base, Serializable):
    __tablename__ = 'compkeys'

    left_id = sa.Column(sa.Integer, primary_key=True)
    right_id = sa.Column(sa.Integer, primary_key=True)

    name = sa.Column(sa.UnicodeText)


class Page(Base, Serializable):
    __tablename__ = 'pages'

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    parent_id = sa.Column(sa.Integer, sa.ForeignKey('pages.id',
        onupdate='cascade', ondelete='cascade'))

    title = sa.Column(sa.Unicode(100))
    content = sa.Column(sa.UnicodeText)

    children = relationship('Page',
            backref=backref('parent', remote_side=id, uselist=False))






'''Tests for the populate() method.'''

import pytest
from sqlalchemy import select
from .models import *


def test_populate(session):
    '''Basic field population'''

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = {
            'firstname': 'The',
            'lastname': 'Test',
            }

    user.populate(data)

    assert user.firstname == 'The' and user.lastname == 'Test'
    session.rollback()

def test_populate_with_none(session):
    '''Basic field population with none'''

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = {
            'firstname': None,
            'lastname': 'Test',
            }

    user.populate(data)

    assert user.firstname is None and user.lastname == 'Test'
    session.rollback()


def test_populate_none_fail(session):
    '''Basic field population with none part 2: failing on purpose'''

    user = User()

    data = {
            'id': None,
            'firstname': None,
            'lastname': 'Test',
            }

    with pytest.raises(TypeError):
        user.populate(data, strict_typing=True)

    session.rollback()

def test_populate_none_bypass(session):
    '''Basic field population with none part 2: skipping the whole thing'''

    user = User()

    data = {
            'id': None,
            'firstname': None,
            'lastname': 'Test',
            }

    user.populate(data, strict_typing=False)
    assert user.id is None and user.firstname is None and user.lastname == 'Test'

    session.rollback()

def test_populate_date(session):
    '''Date field population'''

    from datetime import date

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = {
            'birthdate': '2016-02-18',
            }

    user.populate(data)

    expected = date(2016, 2, 18)


    assert user.birthdate == expected
    session.rollback()


def test_populate_datetime(session):
    '''Datetime field population'''

    from datetime import datetime

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = {
            'created': '2016-02-18 12:34:56'
            }
    user.populate(data)

    expected = datetime(2016, 2, 18, 12, 34, 56)

    assert user.created == expected
    session.rollback()


def test_populate_time(session):
    '''Time field population'''

    from datetime import time

    user = session.scalars(
        select(User).where(User.id == 1)
    ).one()

    data = {
            'alarm': '12:34:56'
            }
    user.populate(data)

    expected = time(12, 34, 56)

    assert user.alarm == expected
    session.rollback()


def test_populate_validation(session):
    '''Validation Helper'''

    data = {
            'username': 'notvalidated',
            'password': 'abcd',
            'smalltext': '12345678',
            }

    u = User()
    errors = u.populate(data, swallow_exceptions=True)

    expected = 2

    assert len(errors) == expected
    session.rollback()


def test_populate_validation2(session):
    '''Validation without exception eating'''

    data = {
            'username': 'notvalidated',
            'password': 'abcd',
            'smalltext': '12345678',
            }

    err = None
    u = User()
    try:
        u.populate(data)
    except ValidationError as ex:
        err = ex

    # For older pythons, we may not know what gets thrown first
    fields = ['password', 'smalltext']

    assert err.args[0] in fields
    session.rollback()


def test_populate_ignore_relationships(session):
    ''' Make sure relationships get ignored '''

    data = {
        'username': 'notvalidated',
        'profile': {
            'somefield': 'OK',
        },
    }

    u = User()
    u.populate(data)

    assert u.profile is None
    session.rollback()


def test_populate_strict(session):
    ''' Test strict typing on populate '''

    data = {
        'username': 'ok',
        'smalltext': 12244,
    }

    u = User()
    field = None

    try:
        u.populate(data, strict_typing=True)
    except TypeError as ex:
        field = ex.args[0]

    assert field == 'smalltext'
    session.rollback()

def test_populate_strict2(session):
    ''' Test strict typing on populate '''

    data = {
        'username': 'ok',
        'smalltext': 12244,
        'password': 'abcd',
    }

    u = User()

    errors = u.populate(data, swallow_exceptions=True, strict_typing=True)
    expected = 2

    assert len(errors) == expected
    session.rollback()



from typing import List, Optional
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from sqlalchemy import (
    Integer,
    PickleType,
    Text,
    Date,
    DateTime,
    String,
    ForeignKey,
    Numeric,
    Float,
    Boolean,
    LargeBinary,
    Enum,
    Interval,
    Identity,
    Table,
    Column,
)
from sqlalchemy.orm import (
    relationship, validates, DeclarativeBase,
    mapped_column, Mapped,
    sessionmaker, scoped_session,
)


from serialalchemy import (
    Serializable, serializable_property,
    serializable_hybrid
)

class Base(DeclarativeBase):
    pass

session_factory = sessionmaker()
Session = scoped_session(session_factory)
# Base.query = Session.query_property()


users_groups = Table('users_groups', Base.metadata,
        Column('user_id', Integer, ForeignKey('users.id',
            onupdate='cascade', ondelete='cascade'), primary_key=True),
        Column('group_id', Integer, ForeignKey('groups.id',
            onupdate='cascade', ondelete='cascade'), primary_key=True)
    )


class ValidationError(Exception):
    pass


class User(Base, Serializable):
    __tablename__ = 'users'

    id: Mapped[int] = mapped_column(Identity(), primary_key=True)

    username: Mapped[Optional[str]] = mapped_column(Text)
    password: Mapped[Optional[str]] = mapped_column(Text)

    firstname: Mapped[Optional[str]] = mapped_column(String(100))
    lastname: Mapped[Optional[str]] = mapped_column(String(100))
    birthdate: Mapped[Optional[date]]

    alarm: Mapped[Optional[time]]
    pickled: Mapped[Optional[bytes]] = mapped_column(PickleType)
    number: Mapped[Optional[Decimal]]
    flt: Mapped[Optional[float]]
    something: Mapped[Optional[bool]]
    blob: Mapped[Optional[bytes]]
    always_excluded: Mapped[Optional[str]] = mapped_column(Text, info={'serializable': False})

    selections: Mapped[Optional[str]] = mapped_column(Enum('one', 'two', 'three'))
    interval: Mapped[Optional[timedelta]]
    smalltext: Mapped[Optional[str]] = mapped_column(String(5))

    created: Mapped[Optional[datetime]] = mapped_column(DateTime(True))

    profile: Mapped["Profile"] = relationship(back_populates="user")
    addresses: Mapped[List["Address"]] = relationship(back_populates="user")
    groups: Mapped[List["Group"]] = relationship(secondary=users_groups, back_populates="users")


    @serializable_property
    def two(self):
        return 1 + 1

    @property
    def three(self):
        return 3

    @validates('smalltext')
    def valid_len(self, key, value):
        if len(value) > 5:
            raise ValidationError(key, 'Invalid length')
        return value

    @validates('password')
    def pw_len(self, key, value):
        if len(value) < 8:
            raise ValidationError(key, 'Password is too short')
        return value

    @serializable_hybrid
    def four(self):
        return 4


class Profile(Base, Serializable):
    __tablename__ = 'profiles'

    id: Mapped[int] = mapped_column(Identity(), primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey('users.id',
                                         onupdate='cascade',
                                         ondelete='cascade'))

    somefield: Mapped[str] = mapped_column(Text)

    user: Mapped["User"] = relationship('User', back_populates='profile', uselist=False)


class Address(Base, Serializable):
    __tablename__ = 'addresses'

    id: Mapped[int] = mapped_column(Identity(), primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey('users.id',
        onupdate='cascade', ondelete='cascade'))

    address: Mapped[Optional[str]] = mapped_column(Text)
    user: Mapped["User"] = relationship(back_populates="addresses")



class Group(Base, Serializable):
    __tablename__ = 'groups'

    id: Mapped[int] = mapped_column(Identity(), primary_key=True)

    name: Mapped[Optional[str]] = mapped_column(Text)

    users: Mapped[List["User"]] = relationship(secondary=users_groups, back_populates="groups")


class Compkey(Base, Serializable):
    __tablename__ = 'compkeys'

    left_id: Mapped[int] = mapped_column(primary_key=True)
    right_id: Mapped[int] = mapped_column(primary_key=True)

    name: Mapped[str] = mapped_column(Text)



class Page(Base, Serializable):
    __tablename__ = 'pages'

    id: Mapped[int] = mapped_column(Identity(), primary_key=True)
    parent_id = mapped_column(ForeignKey('pages.id',
        onupdate='cascade', ondelete='cascade'))

    title: Mapped[str] = mapped_column(Text)
    content: Mapped[str] = mapped_column(Text)

    children: Mapped[List['Page']] = relationship('Page', back_populates='parent')
    parent: Mapped['Page'] = relationship('Page', back_populates='children',
                          remote_side=[id])

